import subprocess as sp

from dataclasses import dataclass
from pydantic import BaseModel
import xml.etree.ElementTree as ET
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
from typing import Callable


@dataclass
class Shell:
    shell: bool = True
    capture_output: bool = True
    run_factory: Callable[..., sp.CompletedProcess] = sp.run

    def run(self, command: str) -> str:
        return self._process_output(
            self.run_factory(
                command, shell=self.shell, capture_output=self.capture_output
            )
        )

    def _process_output(self, output: sp.CompletedProcess):
        if output.returncode == 0:
            return str(output.stdout, encoding="utf-8")
        else:
            return "An error occurred:\n\n" + str(output.stderr, encoding="utf-8")


class XmlFormatter:
    def __init__(self):
        pass

    def to_xml(self, model: BaseModel):
        root_tag = model.__class__.__name__.lower()
        xml = ET.Element(root_tag)
        for k, v in model.model_dump().items():
            sub = ET.SubElement(xml, k.lower())
            safe_text = escape(str(v), entities={"'": "&apos;", '"': "&quot;"})
            sub.text = safe_text
        xml_str = ET.tostring(xml, encoding="unicode")
        return (
            parseString(xml_str).toprettyxml().replace('<?xml version="1.0" ?>\n', "")
        )



from workflows.events import (
    StartEvent,
    Event,
    InputRequiredEvent,
    HumanResponseEvent,
    StopEvent,
)
from pydantic import Field, model_validator
from typing import Optional
from typing_extensions import Self


class MessageEvent(StartEvent):
    message: str


class ProgressEvent(Event):
    msg: str


class GitOrGhEvent(Event):
    main_command: str
    command_information: str


class CommandConstructedEvent(Event):
    command: Optional[str] = Field(default=None)
    subcommand: str
    options: str
    info: str


class CommandExplanationEvent(InputRequiredEvent):
    command: str
    explanation: str


class HumanFeedbackEvent(HumanResponseEvent):
    approved: bool
    feedback: str


class ExecutedEvent(StopEvent):
    is_error: Optional[bool] = Field(default=None)
    output: str

    @model_validator(mode="after")
    def validate_error(self) -> Self:
        if "An error occurred:\n\n" in self.output and not self.is_error:
            self.is_error = True
        elif "An error occurred:\n\n" not in self.output and self.is_error:
            self.is_error = False
        return self



from utils import Shell
from typing import Literal

sh = Shell()


def get_command_info(command: Literal["gh", "git"]) -> str:
    if command == "gh":
        return sh.run("gh help")
    else:
        return sh.run("git --help")


def get_subcommand_info(main_command: Literal["gh", "git"], subcommand: str) -> str:
    return sh.run(f"{main_command} {subcommand} --help")



from src.gut.workflow import GutWorkflow
from src.gut.workflow.events import ExecutedEvent
from workflows import Workflow


def test_workflow_simple() -> None:
    wf = GutWorkflow(timeout=600)
    assert isinstance(wf, Workflow)
    assert wf._timeout == 600


def test_executed_event() -> None:
    ev = ExecutedEvent(output="An error occurred:\n\nError")
    assert ev.is_error
    ev = ExecutedEvent(output="A correct output", is_error=True)
    assert not ev.is_error



import pytest
import xml.etree.ElementTree as ET

from subprocess import CompletedProcess
from xml.dom.minidom import parseString
from pydantic import BaseModel
from src.gut.utils import Shell, XmlFormatter


def mock_run_factory(
    command: str, shell: bool, capture_output: bool
) -> CompletedProcess:
    if command != "git hello":
        return CompletedProcess(
            args=command, returncode=0, stdout=b"hello world", stderr=b""
        )
    return CompletedProcess(
        args=command, returncode=1, stdout=b"", stderr=b"Command is not allowed"
    )


@pytest.fixture
def xml_string_from_basemodel() -> str:
    model = ET.Element("person")
    age = ET.SubElement(model, "age")
    age.text = "30"
    name = ET.SubElement(model, "name")
    name.text = "John Doe"
    xml_str = ET.tostring(model, encoding="unicode")
    return parseString(xml_str).toprettyxml().replace('<?xml version="1.0" ?>\n', "")  # noqa: S318


@pytest.fixture
def starting_value() -> BaseModel:
    class Person(BaseModel):
        age: int
        name: str

    return Person(age=30, name="John Doe")


@pytest.fixture
def sh() -> Shell:
    return Shell(run_factory=mock_run_factory)


def test_shell(sh: Shell):
    assert sh.run("git --help") == "hello world"
    assert sh.run("gh repo") == "hello world"
    assert "An error occurred:\n\n" in sh.run("git hello")


def test_xml(xml_string_from_basemodel: str, starting_value: BaseModel):
    xml = XmlFormatter()
    assert xml.to_xml(starting_value) == xml_string_from_basemodel



import sys
import os
from workflows import Workflow, Context, step
from workflows.resource import Resource
from llama_index.core.llms import LLM, ChatMessage
from typing import Annotated, Union
from pydantic import BaseModel, Field

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from utils import XmlFormatter, Shell
from models import ShellCommand, GitCommand, GhCommand, CommandToExecute
from .events import (
    MessageEvent,
    GitOrGhEvent,
    CommandConstructedEvent,
    CommandExplanationEvent,
    HumanFeedbackEvent,
    ExecutedEvent,
    ProgressEvent,
)
from shell import get_command_info, get_subcommand_info
from .resources import get_shell, get_xml_formatter, get_llm


class WorkflowState(BaseModel):
    user_message: str = Field(default_factory=str)
    main_command: str = Field(default_factory=str)
    command: str = Field(default_factory=str)
    subcommand: str = Field(default_factory=str)
    options: str = Field(default_factory=str)


class GutWorkflow(Workflow):
    @step
    async def choose_command(
        self,
        ev: MessageEvent,
        ctx: Context[WorkflowState],
        llm: Annotated[LLM, Resource(get_llm)],
    ) -> GitOrGhEvent:
        messages = [
            ChatMessage(content=ev.message, role="user"),
            ChatMessage(
                content="Based on my previous message, can you choose, between git and gh, what command would better suit my needs?",
                role="user",
            ),
        ]
        sllm = llm.as_structured_llm(ShellCommand)
        response = await sllm.achat(messages)
        output = ShellCommand.model_validate_json(response.message.content)
        state = await ctx.store.get_state()
        state.user_message = ev.message
        state.main_command = output.command
        await ctx.store.set_state(state)
        ctx.write_event_to_stream(
            ProgressEvent(msg=f"Main command chosen: {output.command}")
        )
        return GitOrGhEvent(
            main_command=output.command,
            command_information=get_command_info(output.command),
        )

    @step
    async def build_command(
        self,
        ev: GitOrGhEvent,
        ctx: Context[WorkflowState],
        llm: Annotated[LLM, Resource(get_llm)],
        formatter: Annotated[XmlFormatter, Resource(get_xml_formatter)],
    ) -> CommandConstructedEvent:
        class Command(BaseModel):
            main_command: str
            command_infomation: str
            user_instructions: str

        state = await ctx.store.get_state()
        mod = Command(
            main_command=ev.main_command,
            command_infomation=ev.command_information,
            user_instructions=state.user_message,
        )
        messages = [
            ChatMessage(
                content=f"Based on this model:\n\n{formatter.to_xml(mod)}\n\nCan you please output the necessary command/subcommand and options to fulfil the user's instructions?"
            )
        ]
        if ev.main_command == "gh":
            sllm = llm.as_structured_llm(GhCommand)
            response = await sllm.achat(messages)
            output = GhCommand.model_validate_json(response.message.content)
            state.command = output.command
            state.subcommand = output.subcommand
            state.options = output.options
        else:
            sllm = llm.as_structured_llm(GitCommand)
            response = await sllm.achat(messages)
            output = GitCommand.model_validate_json(response.message.content)
            state.command = output.subcommand
            state.options = output.options
        await ctx.store.set_state(state)
        ctx.write_event_to_stream(
            ProgressEvent(
                msg=f"Produced command: {state.main_command} {state.command} {state.subcommand} {state.options}"
            )
        )
        return CommandConstructedEvent(
            **output.model_dump(),
            info=get_subcommand_info(
                main_command=ev.main_command, subcommand=state.command
            ),
        )

    @step
    async def explain_command(
        self,
        ev: CommandConstructedEvent,
        ctx: Context[WorkflowState],
        llm: Annotated[LLM, Resource(get_llm)],
        formatter: Annotated[XmlFormatter, Resource(get_xml_formatter)],
    ) -> CommandExplanationEvent:
        state = await ctx.store.get_state()
        command = (
            state.main_command
            + " "
            + state.command
            + " "
            + state.subcommand
            + " "
            + state.options
        )
        command = command.strip()
        command = command.replace("  ", " ")

        class CommandToExplain(BaseModel):
            command: str
            info: str
            user_instructions: str

        mod = CommandToExplain(
            command=command, info=ev.info, user_instructions=state.user_message
        )
        try:
            messages = [
                ChatMessage(
                    content=f"Based on this model:\n\n{formatter.to_xml(mod)}\n\nCan you please explain the command and how it fulfils the user's instructions?"
                )
            ]
        except Exception:
            messages = [
                ChatMessage(
                    content=f"Based on this JSOn:\n\n```json\n{mod.model_dump_json(indent=4)}\n```\n\nCan you please explain the command and how it fulfils the user's instructions?"
                )
            ]
        sllm = llm.as_structured_llm(CommandToExecute)
        response = await sllm.achat(messages)
        output = CommandToExecute.model_validate_json(response.message.content)
        ctx.write_event_to_stream(
            ProgressEvent(msg=f"Explanation for the command: {output.explanation}")
        )
        return CommandExplanationEvent(command=command, explanation=output.explanation)

    @step
    async def execute_command(
        self,
        ev: HumanFeedbackEvent,
        ctx: Context[WorkflowState],
        sh: Annotated[Shell, Resource(get_shell)],
    ) -> Union[MessageEvent, ExecutedEvent]:
        state = await ctx.store.get_state()
        command = (
            state.main_command
            + " "
            + state.command
            + " "
            + state.subcommand
            + " "
            + state.options
        )
        command = command.strip()
        command = command.replace("  ", " ")
        if ev.approved:
            out = sh.run(command)
            return ExecutedEvent(output=out, is_error="An error occurred\n\n:" in out)
        else:
            message = f"My first instructions were: {state.user_message} and you generated: `{command}`. Now I am asking it to re-generate the command with this feedback: {ev.feedback}"
            ctx.write_event_to_stream(MessageEvent(message=message))
            return MessageEvent(message=message)



from pydantic import BaseModel, Field
from typing import Literal


class ShellCommand(BaseModel):
    command: Literal["git", "gh"] = Field(
        description="Choose a command as the base for the execution, either git or gh"
    )


class GitCommand(BaseModel):
    subcommand: Literal[
        # Start a working area
        "clone",
        "init",
        # Work on the current change
        "add",
        "mv",
        "restore",
        "rm",
        # Examine the history and state
        "bisect",
        "diff",
        "grep",
        "log",
        "show",
        "status",
        # Grow, mark, and tweak your common history
        "backfill",
        "branch",
        "commit",
        "merge",
        "rebase",
        "reset",
        "switch",
        "tag",
        "checkout",
        # Collaborate
        "fetch",
        "pull",
        "push",
    ] = Field(
        description="Subcommand to use",
    )
    options: str = Field(
        description="Options for the subcommand.",
        examples=["origin feature-branch", "-m 'commit message'"],
    )


class GhCommand(BaseModel):
    command: Literal[
        # Core Commands
        "auth",
        "browse",
        "codespace",
        "gist",
        "issue",
        "org",
        "pr",
        "project",
        "release",
        "repo",
        # GitHub Actions Commands
        "cache",
        "run",
        "workflow",
        # Alias Commands
        "co",
        # Additional Commands
        "alias",
        "api",
        "attestation",
        "completion",
        "config",
        "extension",
        "gpg-key",
        "label",
        "preview",
        "ruleset",
        "search",
        "secret",
        "ssh-key",
        "status",
        "variable",
    ]
    subcommand: str = Field(
        description="Subcommand for the main gh command",
        examples=["create", "checkout"],
    )
    options: str = Field(
        description="Options for the command and subcommand combination",
        examples=["--fill", "view", "archive"],
    )


class CommandToExecute(BaseModel):
    command: str = Field(
        description="Command to execute",
        examples=["git commit -m 'first commit'", "gh pr checkout 1937"],
    )
    explanation: str = Field(
        description="Detailed-but-concise explanation of why is the command being executed."
    )



import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from llama_index.llms.groq import Groq
from llama_index.core.llms import LLM
from utils import Shell, XmlFormatter
from dotenv import load_dotenv

load_dotenv()

llm = Groq(api_key=os.getenv("GROQ_API_KEY"), model="llama-3.3-70b-versatile")


def get_shell(*args, **kwargs) -> Shell:
    return Shell()


def get_xml_formatter(*args, **kwargs) -> XmlFormatter:
    return XmlFormatter()


def get_llm(*args, **kwargs) -> LLM:
    return llm



import pytest
from pydantic import ValidationError

from src.gut.models import ShellCommand, GhCommand, GitCommand, CommandToExecute


def test_shell_command():
    sh = ShellCommand(command="git")
    assert sh.command == "git"
    with pytest.raises(ValidationError):
        ShellCommand(comment="gut")


def test_gh_command():
    gh = GhCommand(command="pr", subcommand="checkout", options="1997")
    assert gh.command == "pr"
    assert gh.subcommand == "checkout"
    assert gh.options == "1997"
    with pytest.raises(ValidationError):
        GhCommand(command="hello", subcommand="world", options="--fix")


def test_git_command():
    gitcmd = GitCommand(subcommand="commit", options="-m 'first' commit")
    assert gitcmd.subcommand == "commit"
    assert gitcmd.options == "-m 'first' commit"
    with pytest.raises(ValidationError):
        GitCommand(subcommand="hello", options="--fix")


def test_final_command():
    cmd = CommandToExecute(
        command="git commit -m 'first commit'", explanation="This is a commit command."
    )
    assert cmd.command == "git commit -m 'first commit'"
    assert cmd.explanation == "This is a commit command."
    with pytest.raises(ValidationError):
        CommandToExecute(command=1, explanation="This is a wrong command.")



import asyncio

from rich.console import Console
from rich.table import Table
from .workflow import GutWorkflow
from .workflow.events import (
    MessageEvent,
    ProgressEvent,
    CommandExplanationEvent,
    HumanFeedbackEvent,
    ExecutedEvent,
)


async def run_workflow() -> int:
    wf = GutWorkflow(timeout=600, disable_validation=True)
    cs = Console()
    cs.print(
        "[bold cyan]>[/bold cyan] Welcome to gut - your assistant for everything related to [code]git[/code] and [code]gh[/code]! Before we start, let me ask you a question: are you ready to trust your guts on git? [Yes/No]"
    )
    guts = cs.input("[bold magenta]>[/bold magenta]")
    if guts.lower() == "yes":
        cs.print("[bold cyan]> Fantastic[/bold cyan], let's start!")
    else:
        cs.print("[bold cyan]>[/bold cyan] Well, then, see you next time!")
        return 0
    while True:
        cs.print(
            "[bold cyan]>[/bold cyan] So, what would you like me to do now? (Use q to quit)"
        )
        user_message = cs.input("[bold magenta]>[/bold magenta]")
        if user_message == "q":
            return 0
        handler = wf.run(start_event=MessageEvent(message=user_message))
        with cs.status("[bold green]Working on your request...") as status:
            async for event in handler.stream_events():
                if isinstance(event, ProgressEvent):
                    cs.log(event.msg)
                elif isinstance(event, CommandExplanationEvent):
                    cs.log("Here is the explanation of the command:")
                    cs.log(event.explanation)
                    cs.log("Should I go on with executing this command? [Yes/feedback]")
                    status.stop()
                    hitl = cs.input("[bold magenta]>[/bold magenta]")
                    if hitl.strip().lower() == "yes":
                        handler.ctx.send_event(  # type: ignore[union-attr]
                            HumanFeedbackEvent(
                                approved=True,
                                feedback="",
                            )
                        )
                    else:
                        handler.ctx.send_event(  # type: ignore[union-attr]
                            HumanFeedbackEvent(
                                approved=False,
                                feedback=hitl,
                            )
                        )
        result: ExecutedEvent = await handler
        error = "No Errors" if not result.is_error else "yes"
        output = "No Output Captured" if not result.output else result.output
        table = Table(show_footer=False)
        table.title = "Execution Details"
        table.add_column("Captured Output", justify="center")
        table.add_column("Errors", justify="center")
        table.add_row(
            output,
            error,
        )
        cs.print(table)


def main():
    asyncio.run(run_workflow())
